/*
 * IRC - Internet Relay Chat, ircd/p_servmode.c
 * Copyright (C) 1990 Jarkko Oikarinen and
 *                    University of Oulu, Computing Center
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "channel.h"
#include "client.h"
#include "hash.h"
#include "ircd.h"
#include "ircd_string.h"
#include "msg.h"
#include "numeric.h"
#include "numnicks.h"
#include "patch.h"
#include "send.h"

#include <assert.h>
#include <string.h>


/* --------------------------------------------------------
 * **Daz** m_servmode()
 *      This function basically carries out a MODE change
 *      (or indeed a series of mode changes) as the server
 *      (ie. not the client performing them).  This function
 *      should only really ever be used in cases of
 *      emergency, and serveradmins should ensure that this
 *      command is logged to ensure no abuse takes place.
 *
 *      Since this command is essentially a 'wrapper' for
 *      the function: 'set_mode' - all necessary MODE sanity
 *      checking is automatically carried out by the
 *      wrapped function.
 *
 * - parv[0] - sender
 *   parv[1] - channel
 *   parv[2 - parc] - mode(s)/param(s)
 * --------------------------------------------------------
 */
int m_servmode(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
   int badop, sendts, i = (parc - 2);
   struct Channel *chptr;
   char *ptr;
   char modebuf[MODEBUFLEN], parabuf[MODEBUFLEN];
   char nparabuf[MODEBUFLEN];
   char outtext[(MODEBUFLEN + MODEBUFLEN) + 4];
   char ccmodebuf[MODEBUFLEN];

   /* Option enabled? */
   if (!IsOption(OPER_SERVMODE))
      return 0;

   /* This command requires at least 2 (parc = 4) parameters - a channel
    * and a mode (e.g. "#feds","+n").  Some modes require an additional
    * parameter (e.g op, deop, voice, limit, key, etc).
    */
   if (parc < 3) {
      sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SERVMODE");
      return 0;
   }

   sptr->flags &= ~FLAGS_TS8;
 
   chptr = FindChannel(parv[1]);
   if (!chptr) {
      /* Channel doesn't exist */
      sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], parv[1]);
      return 0;
   }

   /* CLEARCHAN functionality :- If the third (2) argument is '-clearchan', we remove all
    * channel modes that are currently set on channel 'chptr'.
    */
   if (!strCasediff(parv[2],"-clearchan")) {
      *modebuf = *parabuf = '\0';
      channel_modes(sptr, ccmodebuf, parabuf, chptr);	    /* Get the current chan modes */
      if (ptr = strchr(ccmodebuf,'+')) {
         *ptr = '-';					    /* Convert plus to minus - we want to reverse existing modes */

         parv[2] = ccmodebuf;	/* parv[2], originaly '-clearchan' becomes a pointer to our clearchan modebuffer (ccmodebuf), 
                                 * the reverse of the current 'chptr' channel modes.
                                 */

         /* If there is a key in the modebuf, we send the key as our fourth (3) parameter to remove it */
         if (strchr(ccmodebuf,'k')) {
           parv[3] = chptr->mode.key;
           i = (parc - 1);
         }

         /* Assuming this command was issued on channel #feds, which was +smntlk (unlikely I know :)
          * the result after the previous statements had executed would be as follows:
          *     	parv[0]   	parv[1]    	parv[2]    	parv[3]
          * From:    	Durzel		#feds		-clearchan	<NULL>
	  * To:		Durzel		#feds		-smntlk		fedskey
          */	

         /* Now process the modes! */
         *modebuf = *parabuf = '\0';
         sendts = set_mode(&me, sptr, chptr, i, parv + 2, modebuf, parabuf,
            nparabuf, &badop, 1);      
      } else
         assert(0 != ptr);	/* This should never happen, there should always be a '+' in resultant 'modebuf' */
   } else {
      /* ..otherwise it's NORMAL MODE(s) - the final '1' parameter is an indication to the function
       * that this mode is being performed by the server, not the client.
       */
      *modebuf = *parabuf = '\0';
      sendts = set_mode(&me, sptr, chptr, parc - 2, parv + 2, modebuf, parabuf,
         nparabuf, &badop, 1);
   }

   /* If the wrapped 'set_mode' function returns an empty string for 'modebuf', we
    * know that the change has been unsuccessful, therefore we don't need to do
    * any further processing.
    */
   if (strlen(modebuf) <= 1) 
      return 0;
   
   /* Broadcast MODE change */
   if (strlen(modebuf) > (size_t)1) {
      if (badop != 2 && strlen(modebuf) > (size_t)1)
         sendto_channel_butserv(chptr, &me, ":%s MODE %s %s %s",
            me.name, chptr->chname, modebuf, parabuf);
      if (IsLocalChannel(chptr->chname))
         return 0;
 
      if (*modebuf == '\0')
         strcpy(modebuf, "+");

      sendto_highprot_butone(&me, 10, "%s " TOK_MODE " %s %s %s",
         NumServ(&me), chptr->chname, modebuf, nparabuf);
   }
  
  /* Log MODE change (if applicable) */
  if (LogOption(OPER_SERVMODE) && MyUser(sptr)) {
    if (strlen(parabuf) < 1)
       sprintf_irc(outtext, "%s", modebuf);
    else
       sprintf_irc(outtext, "%s %s", modebuf, parabuf);
    LogFeature(sptr, OPER_SERVMODE, "MODE %s %s", chptr->chname, outtext);  
  }
 
  return 0;
}
